from typing import Optional
from warnings import warn

from citrine._serialization import properties
from citrine._serialization.serializable import Serializable
from citrine._session import Session
from citrine.informatics.constraints.constraint import Constraint

__all__ = ['ScalarRangeConstraint']


class ScalarRangeConstraint(Serializable['ScalarRangeConstraint'], Constraint):
    """Represents an inequality constraint on a scalar-valued material attribute.

    Parameters
    ----------
    descriptor_key: str
        the key corresponding to a descriptor
    lower_bound: float
        the minimum value in the range
    upper_bound: float
        the maximum value in the range
    lower_inclusive: bool
        if True, will include the lower bound value in the range (default: true)
    upper_inclusive: bool
        if True, will include the max value in the range (default: true)

    """

    descriptor_key = properties.String('descriptor_key')
    lower_bound = properties.Optional(properties.Float, 'min')
    upper_bound = properties.Optional(properties.Float, 'max')
    lower_inclusive = properties.Boolean('min_inclusive')
    upper_inclusive = properties.Boolean('max_inclusive')
    typ = properties.String('type', default='ScalarRange')

    def __init__(self,
                 descriptor_key: str,
                 lower_bound: Optional[float] = None,
                 upper_bound: Optional[float] = None,
                 lower_inclusive: Optional[bool] = None,
                 upper_inclusive: Optional[bool] = None,
                 min: Optional[float] = None,
                 max: Optional[float] = None,
                 min_inclusive: Optional[bool] = None,
                 max_inclusive: Optional[bool] = None,
                 session: Optional[Session] = None):
        if lower_bound is not None and min is not None:
            raise ValueError("Both lower_bound and min were specified.  "
                             "Please only specify lower_bound.")
        if upper_bound is not None and max is not None:
            raise ValueError("Both upper_bound and max were specified.  "
                             "Please only specify upper_bound.")
        if lower_inclusive is not None and min_inclusive is not None:
            raise ValueError("Both lower_inclusive and min_inclusive were specified.  "
                             "Please only specify lower_inclusive.")
        if upper_inclusive is not None and max_inclusive is not None:
            raise ValueError("Both upper_inclusive and max_inclusive were specified.  "
                             "Please only specify upper_inclusive.")
        if min is not None or max is not None:
            msg = "The min/max arguments for ScalarRangeConstraint are deprecated.  " \
                  "Please use lower_bound/upper_bound instead."
            warn(msg, DeprecationWarning)
        if min_inclusive is not None or max_inclusive is not None:
            msg = "The min_inclusive/max_inclusive arguments for ScalarRangeConstraint " \
                  "are deprecated.  " \
                  "Please use lower_inclusive/upper_inclusive instead."
            warn(msg, DeprecationWarning)

        self.descriptor_key = descriptor_key

        if lower_bound is not None:
            self.lower_bound = lower_bound
        else:
            self.lower_bound = min
        if upper_bound is not None:
            self.upper_bound = upper_bound
        else:
            self.upper_bound = max

        # we have to be careful with None and boolean values
        # None or False or True -> True, so that pattern doesn't work
        self.lower_inclusive = True
        if lower_inclusive is not None:
            self.lower_inclusive = lower_inclusive
        elif min_inclusive is not None:
            self.lower_inclusive = min_inclusive

        self.upper_inclusive = True
        if upper_inclusive is not None:
            self.upper_inclusive = upper_inclusive
        elif max_inclusive is not None:
            self.upper_inclusive = max_inclusive

        self.session: Optional[Session] = session

    def __str__(self):
        return '<ScalarRangeConstraint {!r}>'.format(self.descriptor_key)

    @property
    def min(self):
        """[DEPRECATED] Alias for lower_bound."""
        warn("ScalarRangeConstraint.min is deprecated.  Please use lower_bound instead",
             DeprecationWarning)
        return self.lower_bound

    @property
    def max(self):
        """[DEPRECATED] Alias for upper_bound."""
        warn("ScalarRangeConstraint.max is deprecated.  Please use upper_bound instead",
             DeprecationWarning)
        return self.upper_bound

    @property
    def min_inclusive(self):
        """[DEPRECATED] Alias for lower_inclusive."""
        warn("ScalarRangeConstraint.min_inclusive is deprecated.  "
             "Please use lower_inclusive instead",
             DeprecationWarning)
        return self.lower_inclusive

    @property
    def max_inclusive(self):
        """[DEPRECATED] Alias for upper_inclusive."""
        warn("ScalarRangeConstraint.max_inclusive is deprecated.  "
             "Please use upper_inclusive instead",
             DeprecationWarning)
        return self.upper_inclusive



"""Tools for working with Descriptors."""
import warnings
from typing import Type, Optional, Set  # noqa: F401

from citrine._serialization.serializable import Serializable
from citrine._serialization.polymorphic_serializable import PolymorphicSerializable
from citrine._serialization import properties

__all__ = ['Descriptor',
           'RealDescriptor',
           'ChemicalFormulaDescriptor',
           'MolecularStructureDescriptor',
           'CategoricalDescriptor',
           'FormulationDescriptor']


class Descriptor(PolymorphicSerializable['Descriptor']):
    """A Descriptor describes the range of values that a quantity can take on.

    Abstract type that returns the proper type given a serialized dict.
    """

    key = properties.String('descriptor_key')

    @classmethod
    def get_type(cls, data) -> Type[Serializable]:
        """Return the subtype."""
        # Current backend bug PLA-4036 means that some descriptors come back with "category"
        # as type key. This should be resolved soon
        try:
            t = data["type"]
        except KeyError:
            t = data["category"]
        return {
            "Categorical": CategoricalDescriptor,
            "Formulation": FormulationDescriptor,
            "Inorganic": ChemicalFormulaDescriptor,
            "Organic": MolecularStructureDescriptor,
            "Real": RealDescriptor,
        }[t]


class RealDescriptor(Serializable['RealDescriptor'], Descriptor):
    """A descriptor to hold real-valued numbers.

    Parameters
    ----------
    key: str
        the key corresponding to a descriptor
    lower_bound: float
        inclusive lower bound for valid real values
    upper_bound: float
        inclusive upper bound for valid real values

    """

    lower_bound = properties.Float('lower_bound')
    upper_bound = properties.Float('upper_bound')
    units = properties.Optional(properties.String, 'units', default='')
    typ = properties.String('type', default='Real', deserializable=False)

    def __eq__(self, other):
        try:
            attrs = ["key", "lower_bound", "upper_bound", "units", "typ"]
            return all([
                self.__getattribute__(key) == other.__getattribute__(key) for key in attrs
            ])
        except AttributeError:
            return False

    def __init__(self,
                 key: str,
                 lower_bound: float,
                 upper_bound: float,
                 units: Optional[str] = None):
        self.key: str = key
        self.lower_bound: float = lower_bound
        self.upper_bound: float = upper_bound

        if units is None:
            msg = "Default of dimensionless is deprecated; \
            please specify an empty string explicitly."
            warnings.warn(msg, category=DeprecationWarning)
            self.units = ""
        else:
            self.units = units

    def __str__(self):
        return "<RealDescriptor {!r}>".format(self.key)

    def __repr__(self):
        return "RealDescriptor({}, {}, {}, {})".format(
            self.key, self.lower_bound, self.upper_bound, self.units)


class ChemicalFormulaDescriptor(Serializable['ChemicalFormulaDescriptor'], Descriptor):
    """Captures domain-specific context about a stoichiometric chemical formula.

    Parameters
    ----------
    key: str
        the key corresponding to a descriptor

    """

    typ = properties.String('type', default='Inorganic', deserializable=False)

    def __eq__(self, other):
        try:
            attrs = ["key", "typ"]
            return all([
                self.__getattribute__(key) == other.__getattribute__(key) for key in attrs
            ])
        except AttributeError:
            return False

    def __init__(self, key: str):
        self.key: str = key

    def __str__(self):
        return "<ChemicalFormulaDescriptor {!r}>".format(self.key)

    def __repr__(self):
        return "ChemicalFormulaDescriptor(key={})".format(self.key)


class MolecularStructureDescriptor(Serializable['MolecularStructureDescriptor'], Descriptor):
    """
    Material descriptor for an organic molecule.

    Accepts SMILES, IUPAC, and InChI String values.

    Parameters
    ----------
    key: str
        The column header key corresponding to this descriptor

    """

    typ = properties.String('type', default='Organic', deserializable=False)

    def __eq__(self, other):
        try:
            attrs = ["key", "typ"]
            return all([
                self.__getattribute__(key) == other.__getattribute__(key) for key in attrs
            ])
        except AttributeError:
            return False

    def __init__(self, key: str):
        self.key: str = key

    def __str__(self):
        return "<MolecularStructureDescriptor {!r}>".format(self.key)

    def __repr__(self):
        return "MolecularStructureDescriptor(key={})".format(self.key)


class CategoricalDescriptor(Serializable['CategoricalDescriptor'], Descriptor):
    """A descriptor to hold categorical variables.

    An exhaustive list of categorical values may be supplied.

    Parameters
    ----------
    key: str
        the key corresponding to a descriptor
    categories: Set[str]
        possible categories for this descriptor

    """

    typ = properties.String('type', default='Categorical', deserializable=False)
    categories = properties.Set(properties.String, 'descriptor_values')

    def __eq__(self, other):
        try:
            attrs = ["key", "categories", "typ"]
            return all([
                self.__getattribute__(key) == other.__getattribute__(key) for key in attrs
            ])
        except AttributeError:
            return False

    def __init__(self, key: str, categories: Set[str]):
        self.key: str = key
        for category in categories:
            if not isinstance(category, str):
                raise TypeError("All categories must be strings")
        self.categories: Set[str] = categories

    def __str__(self):
        return "<CategoricalDescriptor {!r}>".format(self.key)

    def __repr__(self):
        return "CategoricalDescriptor(key={}, categories={})".format(self.key, self.categories)


class FormulationDescriptor(Serializable['FormulationDescriptor'], Descriptor):
    """A descriptor to hold formulations.

    Parameters
    ----------
    key: str
        the key corresponding to a descriptor

    """

    typ = properties.String('type', default='Formulation', deserializable=False)

    def __eq__(self, other):
        try:
            attrs = ["key", "typ"]
            return all([
                self.__getattribute__(key) == other.__getattribute__(key) for key in attrs
            ])
        except AttributeError:
            return False

    def __init__(self, key: str):
        self.key: str = key

    def __str__(self):
        return "<FormulationDescriptor {!r}>".format(self.key)

    def __repr__(self):
        return "FormulationDescriptor(key={})".format(self.key)



"""Property objects for typed setting and ser/de."""
from abc import abstractmethod
import typing
from typing import Optional
from datetime import datetime
from itertools import chain
import uuid
import arrow

from gemd.enumeration.base_enumeration import BaseEnumeration
from gemd.entity.link_by_uid import LinkByUID
from gemd.entity.dict_serializable import DictSerializable
from citrine._serialization.serializable import Serializable
from citrine._serialization.polymorphic_serializable import PolymorphicSerializable

SerializedType = typing.TypeVar('SerializedType')
DeserializedType = typing.TypeVar('DeserializedType')
SerializedInteger = typing.TypeVar('SerializedInteger', int, str)
SerializedFloat = typing.TypeVar('SerializedFloat', float, str)


class Property(typing.Generic[DeserializedType, SerializedType]):

    def __init__(self,
                 serialization_path: typing.Optional[str] = None,
                 serializable: bool = True,
                 deserializable: bool = True,
                 default: typing.Optional[DeserializedType] = None,
                 override: bool = False
                 ):
        self.serialization_path = serialization_path
        self._key: str = '__' + str(uuid.uuid4())  # Make this object key human readable
        self.serializable: bool = serializable
        self.deserializable: bool = deserializable
        self.default: typing.Optional[DeserializedType] = default
        # Distinguish between no default being provided and the default being None
        self.optional = False
        self.override = override

    @property
    @abstractmethod
    def underlying_types(self) -> typing.Union[DeserializedType, typing.Tuple[DeserializedType]]:
        """Return the python types handled by this property."""

    @property
    @abstractmethod
    def serialized_types(self) -> typing.Union[SerializedType, typing.Tuple[SerializedType]]:
        """Return the types used to serialize this property."""

    def _error_source(self, base_class: type) -> str:
        """Construct a string of the base class name and the parameter that failed."""
        if base_class is not None:
            return 'for {}:{}'.format(base_class.__name__, self.serialization_path)
        elif self.serialization_path:
            return 'for {}'.format(self.serialization_path)
        else:
            return ''

    def serialize(self, value: DeserializedType,
                  base_class: Optional[type] = None) -> SerializedType:
        if not isinstance(value, self.underlying_types):
            base_name = self._error_source(base_class)
            raise ValueError('{} is not one of valid types: {} {}'.format(
                value, self.underlying_types, base_name))
        return self._serialize(value)

    def deserialize(self, value: SerializedType,
                    base_class: Optional[type] = None) -> DeserializedType:
        if not isinstance(value, self.serialized_types):
            base_name = self._error_source(base_class)
            raise ValueError('{} is not one of valid types: {} {}'.format(
                value, self.serialized_types, base_name))
        return self._deserialize(value)

    @abstractmethod
    def _serialize(self, value: DeserializedType) -> SerializedType:
        """Perform serialization."""

    @abstractmethod
    def _deserialize(self, value: SerializedType) -> DeserializedType:
        """Perform deserialization."""

    def deserialize_from_dict(self, data: dict) -> DeserializedType:
        value = data
        # `serialization_path` is expected to be a sequence of nested dictionary keys
        fields = self.serialization_path.split('.')
        for field in fields:
            next_value = value.get(field)
            if next_value is None:
                if self.default is None and not self.optional:
                    msg = "Unable to deserialize {} into {}: missing a required field {}".format(
                        data, self.underlying_types, field)
                    raise RuntimeError(msg)
                # This occurs if a `field` is unexpectedly not present in the data dictionary
                # or if its value is null.
                # Use the default value and stop traversing, even if we have not yet reached
                # the last field in the serialization path.
                value = self.serialize(self.default)
                break
            else:
                value = next_value
        base_class = _get_base_class(data, self.serialization_path)
        return self.deserialize(value, base_class=base_class)

    def serialize_to_dict(self, data: dict, value: DeserializedType) -> dict:
        if self.serialization_path is None:
            raise ValueError('No serialization path set!')
        else:
            base_class = _get_base_class(data, self.serialization_path)
            _data = data
            fields = self.serialization_path.split('.')
            for field in fields[:-1]:
                _data = _data.setdefault(field, {})
            _data[fields[-1]] = self.serialize(value, base_class=base_class)
            return data

    def __get__(self, obj, objtype=None) -> DeserializedType:
        """Property getter, deferring to the getter of the parent class, if applicable."""
        base_class = _get_base_class(obj, self.serialization_path)
        if base_class is not None and self.override:
            return getattr(base_class, self.serialization_path).fget(obj)
        else:
            return getattr(obj, self._key, self.default)

    def __set__(self, obj, value: typing.Union[SerializedType, DeserializedType]):
        """Property setter, deferring to the setter of the parent class, if applicable."""
        base_class = _get_base_class(obj, self.serialization_path)
        if issubclass(type(value), self.underlying_types):
            value_to_set = value
        else:
            # if value is not an underlying type, set its deserialized version.
            value_to_set = self.deserialize(value, base_class=base_class)

        if base_class is not None and self.override:
            getattr(base_class, self.serialization_path).fset(obj, value_to_set)
        else:
            setattr(obj, self._key, value_to_set)

    def __str__(self):
        return '<Property {!r}>'.format(self.serialization_path)


class PropertyCollection(Property[DeserializedType, SerializedType]):

    def __set__(self, obj, value: typing.Union[SerializedType, DeserializedType]):
        """
        Property setter for container property types.

        This setter defers to the subclass to implement the `_set_elements` logic
        """
        base_class = _get_base_class(obj, self.serialization_path)
        if issubclass(type(value), self.underlying_types):
            value_to_set = self._set_elements(value)
        else:
            # if value is not an underlying type, set its deserialized version.
            value_to_set = self.deserialize(value, base_class=base_class)

        if base_class is not None and self.override:
            getattr(base_class, self.serialization_path).fset(obj, value_to_set)
        else:
            setattr(obj, self._key, value_to_set)

    @abstractmethod
    def _set_elements(self, value: typing.Union[SerializedType, DeserializedType]):
        """
        Perform any needed underlying element specific deserialization.

        Return the appropriate value to set
        """


def _get_base_class(obj: object, key: str) -> type:
    """
    Return the base class that has key as an attribute, if it exists.

    If there are no base classes with key as an attribute, OR if there are multiple base classes
    with key as an attribute, return None.
    """
    base_classes = obj.__class__.__bases__  # Tuple of all base classes of obj
    try:
        classes_with_key = [base_class for base_class in base_classes if hasattr(base_class, key)]
    except TypeError:
        return None
    if len(classes_with_key) == 1:
        return classes_with_key[0]
    else:
        return None


class Integer(Property[int, SerializedInteger]):

    @property
    def underlying_types(self):
        return int

    @property
    def serialized_types(self):
        return int, str

    def _deserialize(self, value: SerializedInteger) -> int:
        if isinstance(value, bool):
            raise TypeError('value must be a Number, not a boolean.')
        else:
            return int(value)

    def _serialize(self, value: int) -> SerializedInteger:
        if isinstance(value, bool):
            raise TypeError('Boolean cannot be serialized to integer.')
        else:
            return value

    def __str__(self):
        return '<Integer {!r}>'.format(self.serialization_path)


class Float(Property[float, SerializedFloat]):

    @property
    def underlying_types(self):
        return float

    @property
    def serialized_types(self):
        return float, int, str

    @classmethod
    def _deserialize(cls, value: SerializedFloat) -> float:
        if isinstance(value, bool):
            raise TypeError('value must be a Number, not a boolean.')
        else:
            return float(value)

    @classmethod
    def _serialize(cls, value: float) -> SerializedFloat:
        return value

    def __str__(self):
        return '<Float {!r}>'.format(self.serialization_path)


class Raw(Property[typing.Any, typing.Any]):

    @property
    def underlying_types(self):
        return object

    @property
    def serialized_types(self):
        return object

    @classmethod
    def _deserialize(cls, value: typing.Any) -> typing.Any:
        return value

    @classmethod
    def _serialize(cls, value: typing.Any) -> typing.Any:
        return value

    def __str__(self):
        return '<Raw {!r}>'.format(self.serialization_path)


class String(Property[str, str]):

    @property
    def underlying_types(self):
        return str

    @property
    def serialized_types(self):
        return str

    def _deserialize(self, value: str) -> str:
        value = self.default if value is None else value
        if value is None:
            raise ValueError('Value must not be none!')
        return str(value)

    def _serialize(self, value: str) -> str:
        return str(value)

    def __str__(self):
        return '<String {!r}>'.format(self.serialization_path)


class Boolean(Property[bool, bool]):

    @property
    def underlying_types(self):
        return bool

    @property
    def serialized_types(self):
        return bool

    def _deserialize(self, value: str) -> bool:
        return bool(value)

    def _serialize(self, value: str) -> bool:
        return bool(value)

    def __str__(self):
        return '<Boolean {!r}>'.format(self.serialization_path)


class UUID(Property[uuid.UUID, str]):

    @property
    def underlying_types(self):
        return uuid.UUID

    @property
    def serialized_types(self):
        return str

    def _deserialize(self, value: str) -> uuid.UUID:
        return uuid.UUID(value)

    def _serialize(self, value: uuid.UUID) -> str:
        return str(value)


class Datetime(Property[datetime, int]):

    @property
    def underlying_types(self):
        return datetime

    @property
    def serialized_types(self):
        return int, str

    def _deserialize(self, value) -> datetime:
        if isinstance(value, str):
            return arrow.get(value).datetime
        if isinstance(value, int):
            # Backend returns time as ms since epoch, but arrow expects seconds since epoch
            return arrow.get(value / 1000).datetime
        raise TypeError("{} must be an int or a string".format(value))

    def _serialize(self, value: datetime) -> int:
        return int(arrow.get(value).float_timestamp * 1000)


class List(PropertyCollection[list, list]):

    def __init__(self,
                 element_type: typing.Union[Property, typing.Type[Property]],
                 serialization_path: typing.Optional[str] = None,
                 serializable: bool = True,
                 deserializable: bool = True,
                 default: typing.Optional[DeserializedType] = None,
                 override: bool = False):
        super().__init__(serialization_path,
                         serializable,
                         deserializable,
                         default,
                         override)
        self.element_type = element_type if isinstance(element_type, Property) else element_type()

    @property
    def underlying_types(self):
        return list, set, tuple

    @property
    def serialized_types(self):
        return list

    def _deserialize(self, value: list) -> list:
        deserialized = []
        for element in value:
            deserialized.append(self.element_type.deserialize(element))
        return deserialized

    def _serialize(self, value: list) -> list:
        serialized = []
        for element in value:
            serialized.append(self.element_type.serialize(element))
        return serialized

    def _set_elements(self, value):
        elems = []
        for sub_val in value:
            if isinstance(self.element_type, PropertyCollection):
                val_to_append = self.element_type._set_elements(sub_val)
            elif issubclass(type(sub_val), self.element_type.underlying_types):
                val_to_append = sub_val
            else:
                val_to_append = self.element_type.deserialize(sub_val)
            elems.append(val_to_append)
        return elems


class Set(PropertyCollection[set, typing.Iterable]):

    def __init__(self,
                 element_type: typing.Union[Property, typing.Type[Property]],
                 serialization_path: typing.Optional[str] = None,
                 serializable: bool = True,
                 deserializable: bool = True,
                 default: typing.Optional[DeserializedType] = None,
                 override: bool = False):
        super().__init__(serialization_path,
                         serializable,
                         deserializable,
                         default,
                         override)
        self.element_type = element_type if isinstance(element_type, Property) else element_type()

    @property
    def underlying_types(self):
        return set, list, tuple

    @property
    def serialized_types(self):
        return typing.Iterable

    def _deserialize(self, value: typing.Iterable) -> set:
        deserialized = set()
        for element in value:
            deserialized.add(self.element_type.deserialize(element))
        return deserialized

    def _serialize(self, value: set) -> list:
        serialized = list()
        for element in value:
            serialized.append(self.element_type.serialize(element))
        try:
            return sorted(serialized)
        except TypeError:
            return serialized

    def _set_elements(self, value):
        elems = set()
        for sub_val in value:
            if isinstance(self.element_type, PropertyCollection):
                val_to_append = self.element_type._set_elements(sub_val)
            elif issubclass(type(sub_val), self.element_type.underlying_types):
                val_to_append = sub_val
            else:
                val_to_append = self.element_type.deserialize(sub_val)
            elems.add(val_to_append)
        return elems


class Union(Property[typing.Any, typing.Any]):
    """
    One of several possible property types.

    Attempted de/serialization is done in the order in which types are provided in the constructor.
    """

    def __init__(self,
                 element_types: typing.Sequence[typing.Union[Property, typing.Type[Property]]],
                 serialization_path: typing.Optional[str] = None,
                 serializable: bool = True,
                 deserializable: bool = True,
                 default: typing.Optional[DeserializedType] = None,
                 override: bool = False):
        super().__init__(serialization_path,
                         serializable,
                         deserializable,
                         default,
                         override)
        if not isinstance(element_types, typing.Iterable):
            raise ValueError("element types must be iterable: {}".format(element_types))
        self.element_types: typing.List[Property, ...] = \
            [el if isinstance(el, Property) else el() for el in element_types]

    @property
    def underlying_types(self):
        all_underlying_types = [prop.underlying_types for prop in self.element_types]
        return tuple(set(chain(*[typ if isinstance(typ, tuple)
                                 else (typ,) for typ in all_underlying_types])))

    @property
    def serialized_types(self):
        all_serialized_types = [prop.serialized_types for prop in self.element_types]
        return tuple(set(chain(*[typ if isinstance(typ, tuple)
                                 else (typ,) for typ in all_serialized_types])))

    def _serialize(self, value: typing.Any) -> typing.Any:
        for prop in self.element_types:
            try:
                return prop.serialize(value)
            except ValueError:
                pass
        raise RuntimeError("An unexpected error occurred while trying to serialize {} to one "
                           "of the following types: {}.".format(value, self.serialized_types))

    def _deserialize(self, value: typing.Any) -> typing.Any:
        for prop in self.element_types:
            try:
                return prop.deserialize(value)
            except ValueError:
                pass
        # this is a failsafe that shouldn't actually ever get hit, hence no cover
        msg = "An unexpected error occurred while trying to deserialize {} to one of the " \
              "following types: {}.".format(value, self.underlying_types)  # pragma: no cover
        raise RuntimeError(msg)  # pragma: no cover


class SpecifiedMixedList(PropertyCollection[list, list]):
    """A finite list in which the type of each entry is specified."""

    def __init__(self,
                 element_types: typing.Sequence[typing.Union[Property, typing.Type[Property]]],
                 serialization_path: typing.Optional[str] = None,
                 serializable: bool = True,
                 deserializable: bool = True,
                 default: typing.Optional[DeserializedType] = None,
                 override: bool = False):
        super().__init__(serialization_path,
                         serializable,
                         deserializable,
                         default,
                         override)
        if not isinstance(element_types, list):
            raise ValueError("element types must be a list: {}".format(element_types))
        self.element_types: typing.List[Property, ...] = \
            [el if isinstance(el, Property) else el() for el in element_types]

    @property
    def underlying_types(self):
        return list, tuple

    @property
    def serialized_types(self):
        return list

    def _deserialize(self, value: list) -> tuple:
        if len(value) > len(self.element_types):
            raise ValueError("Cannot deserialize value {}, as it has more elements "
                             "than expected for list {}".format(value, self.element_types))
        deserialized = []
        for element, element_type in zip(value, self.element_types):
            deserialized.append(element_type.deserialize(element))

        # If there are more element types than elements, append default values
        for element_type in self.element_types[len(value):]:
            deserialized.append(element_type.default)

        return deserialized

    def _serialize(self, value: tuple) -> list:
        if len(value) > len(self.element_types):
            raise ValueError("Cannot serialize value {}, as it has more elements "
                             "than expected for list {}".format(value, self.element_types))
        serialized = []
        for element, element_type in zip(value, self.element_types):
            serialized.append(element_type.serialize(element))

        # If there are more element types than elements, append serialized default values
        for element_type in self.element_types[len(value):]:
            serialized.append(element_type.serialize(element_type.default))

        return serialized

    def _set_elements(self, value):
        elems = []
        if len(value) > len(self.element_types):
            raise ValueError("Cannot serialize value {}, as it has more elements "
                             "than expected for list {}".format(value, self.element_types))
        for element, element_type in zip(value, self.element_types):
            if isinstance(element_type, PropertyCollection):
                val_to_append = element_type._set_elements(element)
            elif issubclass(type(element), element_type.underlying_types):
                val_to_append = element
            else:
                val_to_append = element_type.deserialize(element)
            elems.append(val_to_append)

        # If there are more element types than elements, append serialized default values
        for element_type in self.element_types[len(value):]:
            elems.append(element_type.default)

        return elems


class Enumeration(Property[BaseEnumeration, str]):

    def __init__(self,
                 klass: typing.Type[typing.Any],
                 serialization_path: typing.Optional[str] = None,
                 serializable: bool = True,
                 deserializable: bool = True,
                 default: typing.Optional[DeserializedType] = None,
                 override: bool = False):
        super().__init__(serialization_path,
                         serializable,
                         deserializable,
                         default,
                         override)
        self.klass = klass

    @property
    def underlying_types(self):
        return self.klass

    @property
    def serialized_types(self):
        return str

    def _deserialize(self, value: str) -> uuid.UUID:
        return self.klass.get_enum(value)

    def _serialize(self, value: typing.Any) -> str:
        return self.klass.get_value(value)


class Object(PropertyCollection[typing.Any, dict]):

    def __init__(self,
                 klass: typing.Type[typing.Any],
                 serialization_path: typing.Optional[str] = None,
                 serializable: bool = True,
                 deserializable: bool = True,
                 default: typing.Optional[DeserializedType] = None,
                 override: bool = False):
        super().__init__(serialization_path,
                         serializable,
                         deserializable,
                         default,
                         override)
        self.klass = klass
        # We need to use __dict__ here because other access methods will invoke __get__
        # Start with the fields from the parent classes, overwriting with any newer classes
        self.fields = {k: v
                       for x in self.klass.__bases__
                       for k, v in x.__dict__.items() if isinstance(v, Property)}
        self.fields.update(
            {k: v for k, v in self.klass.__dict__.items() if isinstance(v, Property)})
        self.polymorphic = "get_type" in self.klass.__dict__ and\
                           issubclass(self.klass, PolymorphicSerializable)

    @property
    def underlying_types(self):
        return self.klass

    @property
    def serialized_types(self):
        return dict

    def _deserialize(self, data: dict) -> typing.Any:
        if self.polymorphic:
            return self.klass.get_type(data).build(data)
        if not self.fields:
            # Maybe there are no fields because we hit a gemd-python class
            if issubclass(self.klass, DictSerializable):
                return DictSerializable.build(data)
            raise AttributeError("Tried to deserialize to {!r}, which has no fields and is not an"
                                 " explicitly serializable class".format(self.klass))

        instance = self.klass.__new__(self.klass, {})
        for property_name, field in self.fields.items():
            if field.deserializable:
                value = field.deserialize_from_dict(data)
                setattr(instance, property_name, value)
        return instance

    def _serialize(self, obj: typing.Any) -> dict:
        serialized = {}
        if type(obj) != self.klass and isinstance(obj, Serializable):
            # If the object class doesn't match this one, then it is a subclass
            # that may have more fields, so defer to them by calling the dump method
            # it must have as a Serializable
            return obj.dump()
        if not self.fields:
            # There are two types of objects that we expect to not have fields.
            # One is a PolymorphicSerializable, which is handled above.
            # The other possibility is that obj is a gemd object that is not reproduced in
            # citrine-python (attribute, value, bounds, etc.). These are all DictSerializable,
            # and have a dump() method that uses the gemd json encoder client.
            try:
                return obj.dump()
            except AttributeError:
                raise AttributeError("Tried to serialize object {!r} of type {}, which has "
                                     "neither fields not a dump() method.".format(obj, type(obj)))
        for property_name, field in self.fields.items():
            if field.serializable:
                value = getattr(obj, property_name)
                serialized = field.serialize_to_dict(serialized, value)
        return serialized

    def __str__(self):
        return '<Object[{}] {!r}>'.format(self.klass.__name__, self.serialization_path)

    def _set_elements(self, value):
        if issubclass(type(value), self.klass):
            return value
        else:
            return self.deserialize(value)


class LinkOrElse(PropertyCollection[typing.Union[Serializable, LinkByUID], dict]):
    """
    A property that can either be a serializable object with IDs or a LinkByUID object.

    Serialization is done by converting the object to a dict. If it is a Serializable object
    it has a dump() method. LinkByUID has the as_dict() method.

    Deserialization to a LinkByUID is easy--it is constructed using value. Deserialization to
    a Serializable object is more difficult because this class does not know what object it is
    supposed to deserialize to (passing that information results in a circular import, because the
    objects that need to be deserialized in this use case have their own Property serde schema).
    Instead, we rely on the fact that all of the Serializable classes with serde schema
    (data concepts objects, templates, etc.) know how to deserialize themselves through a
    build() method that doesn't call _deserialize().

    This is currently tied to LinkByUID in gemd, but could be modified to work with a
    generic Link object.
    """

    def __init__(self,
                 serialization_path: typing.Optional[str] = None,
                 serializable: bool = True,
                 deserializable: bool = True,
                 default: typing.Optional[DeserializedType] = None,
                 override: bool = False):
        super().__init__(serialization_path, serializable, deserializable, default, override)

    @property
    def underlying_types(self):
        return Serializable, LinkByUID

    @property
    def serialized_types(self):
        return dict

    def _serialize(self, value: typing.Any) -> dict:
        if isinstance(value, LinkByUID):
            return value.as_dict()
        elif isinstance(value, Serializable):
            return value.dump()

    def _deserialize(self, value: dict):
        if 'type' in value and value['type'] == LinkByUID.typ:
            if 'scope' in value and 'id' in value:
                value.pop('type')
                return LinkByUID(**value)
            else:
                raise ValueError("LinkByUID dictionary must have both scope and id fields")

        raise Exception("Serializable object that is being pointed to must have a self-contained "
                        "build() method that does not call deserialize().")

    def _set_elements(self, value):
        return value


class Optional(PropertyCollection[typing.Optional[typing.Any], typing.Optional[typing.Any]]):

    def __init__(self,
                 prop: typing.Union[Property, typing.Type[Property]],
                 serialization_path: typing.Optional[str] = None,
                 serializable: bool = True,
                 deserializable: bool = True,
                 default: typing.Optional[DeserializedType] = None,
                 override: bool = False
                 ):
        super().__init__(serialization_path,
                         serializable,
                         deserializable,
                         default,
                         override)
        self.prop = prop if isinstance(prop, Property) else prop()
        self.optional = True

    @property
    def underlying_types(self):
        constituent_types = self.prop.underlying_types
        if isinstance(constituent_types, tuple):
            return constituent_types + (type(None),)
        else:
            return constituent_types, type(None)

    @property
    def serialized_types(self):
        constituent_types = self.prop.serialized_types
        if isinstance(constituent_types, tuple):
            return constituent_types + (type(None),)
        else:
            return constituent_types, type(None)

    def _deserialize(self, data: typing.Optional[typing.Any]) -> typing.Optional[typing.Any]:
        return self.prop.deserialize(data) if data is not None else None

    def _serialize(self, obj: typing.Optional[typing.Any]) -> typing.Optional[typing.Any]:
        return self.prop.serialize(obj) if obj is not None else None

    def __str__(self):
        return '<Optional[{}] {!r}>'.format(self.prop, self.serialization_path)

    def _set_elements(self, value):
        elem = None
        if value is not None:
            if isinstance(self.prop, PropertyCollection):
                elem = self.prop._set_elements(value)
            else:
                elem = value
        return elem


class Mapping(PropertyCollection[dict, dict]):
    """
    Serialization of a Mapping.

    Serialization is done by converting the map to a dict by default, serializing the
    keys and the values.

    If the optional parameter `ser_as_list_of_pairs` is set to True, serialization is done
    by converting the map to a list of key value pairs. Deserialization expects a list of
    key value pairs and converts them to a dict.
    """

    def __init__(self,
                 keys_type: typing.Union[Property, typing.Type[Property]],
                 values_type: typing.Union[Property, typing.Type[Property]],
                 serialization_path: typing.Optional[str] = None,
                 serializable: bool = True,
                 deserializable: bool = True,
                 default: typing.Optional[dict] = None,
                 override: bool = False,
                 ser_as_list_of_pairs: bool = False):
        super().__init__(serialization_path,
                         serializable,
                         deserializable,
                         default,
                         override)

        self.keys_type = keys_type if isinstance(keys_type, Property) else keys_type()
        self.values_type = values_type if isinstance(values_type, Property) else values_type()
        self.ser_as_list_of_pairs = ser_as_list_of_pairs

    @property
    def underlying_types(self):
        return dict

    @property
    def serialized_types(self):
        if self.ser_as_list_of_pairs:
            return list
        else:
            return dict

    def _deserialize(self, value: typing.Union[dict, list]) -> dict:
        deserialized = dict()

        if type(value) == list:
            for pair in value:
                deserialized_key = self.keys_type.deserialize(pair[0])
                deserialized_value = self.values_type.deserialize(pair[1])
                deserialized[deserialized_key] = deserialized_value
            return deserialized

        for key, value in value.items():
            deserialized_key = self.keys_type.deserialize(key)
            deserialized_value = self.values_type.deserialize(value)
            deserialized[deserialized_key] = deserialized_value
        return deserialized

    def _serialize(self, value: dict) -> typing.Union[dict, list]:
        if self.ser_as_list_of_pairs:
            serialized = []
            for key, value in value.items():
                serialized_key = self.keys_type.serialize(key)
                serialized_value = self.values_type.serialize(value)
                serialized.append((serialized_key, serialized_value))
            return serialized

        serialized = dict()
        for key, value in value.items():
            serialized_key = self.keys_type.serialize(key)
            serialized_value = self.values_type.serialize(value)
            serialized[serialized_key] = serialized_value
        return serialized

    def _set_elements(self, value):
        elems = dict()
        for key, val in value.items():
            if isinstance(self.values_type, PropertyCollection):
                deserialized_value = self.values_type._set_elements(val)
            elif issubclass(type(val), self.values_type.underlying_types):
                deserialized_value = val
            else:
                deserialized_value = self.values_type.deserialize(val)

            if isinstance(self.keys_type, PropertyCollection):
                deserialized_key = self.keys_type._set_elements(key)
            elif issubclass(type(key), self.keys_type.underlying_types):
                deserialized_key = key
            else:
                deserialized_key = self.keys_type.deserialize(key)

            elems[deserialized_key] = deserialized_value
        return elems



from typing import List, Union, Optional
from uuid import UUID
from copy import deepcopy

from citrine._rest.resource import Resource, ResourceTypeEnum
from citrine._serialization import properties
from citrine._session import Session
from citrine.informatics.design_spaces.design_space import DesignSpace
from citrine.informatics.dimensions import Dimension
from citrine._rest.ai_resource_metadata import AIResourceMetadata

__all__ = ['ProductDesignSpace']


class ProductDesignSpace(Resource['ProductDesignSpace'], DesignSpace, AIResourceMetadata):
    """A Cartesian product of design spaces.

    Factors can be other design spaces and/or univariate dimensions.

    Parameters
    ----------
    name:str
        the name of the design space
    description:str
        the description of the design space
    subspaces: List[Union[UUID, DesignSpace]]
        the list of subspaces to combine, either design spaces defined in-line
        or UUIDs that reference design spaces on the platform
    dimensions: list[Dimension]
        univariate dimensions that are factors of the design space; can be enumerated or continuous

    """

    _response_key = None
    _resource_type = ResourceTypeEnum.MODULE

    subspaces = properties.List(properties.Union(
        [properties.UUID, properties.Object(DesignSpace)]
    ), 'config.subspaces', default=[])
    dimensions = properties.Optional(
        properties.List(properties.Object(Dimension)), 'config.dimensions'
    )
    typ = properties.String('config.type', default='ProductDesignSpace', deserializable=False)
    # Product design spaces should not be embedded in other subspaces, hence status is required
    status = properties.String('status', serializable=False)

    # NOTE: These could go here or in _post_dump - it's unclear which is better right now
    module_type = properties.String('module_type', default='DESIGN_SPACE')

    def __init__(self, *,
                 name: str,
                 description: str,
                 subspaces: Optional[List[Union[UUID, DesignSpace]]] = None,
                 dimensions: Optional[List[Dimension]] = None,
                 session: Session = Session()):
        self.name: str = name
        self.description: str = description
        self.subspaces: List[Union[UUID, DesignSpace]] = subspaces or []
        self.dimensions: List[Dimension] = dimensions or []
        self.session: Session = session

    def dump(self) -> dict:
        """Override dump to replace on-platform subspaces with their uids."""
        model_copy = deepcopy(self)
        for i, subspace in enumerate(model_copy.subspaces):
            if isinstance(subspace, DesignSpace) and subspace.uid is not None:
                model_copy.subspaces[i] = subspace.uid
        serialized = properties.Object(ProductDesignSpace).serialize(model_copy)
        return self._post_dump(serialized)

    def _post_dump(self, data: dict) -> dict:
        data['display_name'] = data['config']['name']
        for i, subspace in enumerate(data['config']['subspaces']):
            if isinstance(subspace, dict):
                # embedded design spaces are not modules, so only serialize their config
                data['config']['subspaces'][i] = subspace['config']
        return data

    @classmethod
    def _pre_build(cls, data: dict) -> dict:
        subspaces = data['config'].get('subspaces', [])
        # For each subspace, rename the `instance` key to `config`.
        for i, _ in enumerate(subspaces):
            data['config']['subspaces'][i]['config'] = \
                data['config']['subspaces'][i].pop('instance')
        return data

    def __str__(self):
        return '<ProductDesignSpace {!r}>'.format(self.name)



from typing import List, Optional
from warnings import warn

from citrine._serialization import properties
from citrine._serialization.serializable import Serializable
from citrine._session import Session
from citrine.informatics.constraints.constraint import Constraint

__all__ = ['AcceptableCategoriesConstraint', 'CategoricalConstraint']


class AcceptableCategoriesConstraint(Serializable['AcceptableCategoriesConstraint'], Constraint):
    """
    A constraint on a categorical material attribute to be in a set of acceptable values.

    Parameters
    ----------
    descriptor_key: str
        the key corresponding to the associated Categorical descriptor
    acceptable_categories: list[str]
        the names of the acceptable categories to constrain to

    """

    descriptor_key = properties.String('descriptor_key')
    acceptable_categories = properties.List(properties.String(), 'acceptable_classes')
    typ = properties.String('type', default='AcceptableCategoriesConstraint')

    def __init__(self,
                 descriptor_key: str,
                 acceptable_categories: List[str],
                 session: Optional[Session] = None):
        self.descriptor_key = descriptor_key
        self.acceptable_categories = acceptable_categories
        self.session = session

    def __str__(self):
        return '<AcceptableCategoriesConstraint {!r}>'.format(self.descriptor_key)


class CategoricalConstraint(Serializable['CategoricalConstraint'], Constraint):
    """
    [DEPRECATED] A constraint on a categorical material attribute in a set of acceptable values.

    Parameters
    ----------
    descriptor_key: str
        the key corresponding to the associated Categorical descriptor
    acceptable_categories: list[str]
        the names of the acceptable categories to constrain to

    """

    descriptor_key = properties.String('descriptor_key')
    acceptable_categories = properties.List(properties.String(), 'acceptable_classes')
    typ = properties.String('type', default='AcceptableCategoriesConstraint')

    def __init__(self,
                 descriptor_key: str,
                 acceptable_categories: List[str],
                 session: Optional[Session] = None):
        msg = "{this_class} is deprecated. Please use {replacement} instead.".format(
            this_class="CategoricalConstraint",
            replacement=AcceptableCategoriesConstraint.__name__)
        warn(msg, category=DeprecationWarning)
        self.descriptor_key = descriptor_key
        self.acceptable_categories = acceptable_categories
        self.session = session

    def __str__(self):
        return '<CategoricalConstraint {!r}>'.format(self.descriptor_key)



"""Resources that represent both individual and collections of workflow executions."""
from functools import partial
from typing import Optional, Iterable, Union
from uuid import UUID

from citrine._rest.collection import Collection
from citrine._rest.paginator import Paginator
from citrine._rest.pageable import Pageable
from citrine._rest.resource import Resource
from citrine._serialization import properties
from citrine._session import Session
from citrine.informatics.design_candidate import DesignCandidate
from citrine.resources.response import Response
from citrine.informatics.scores import Score
from citrine.informatics.descriptors import Descriptor


class DesignExecution(Resource['DesignExecution'], Pageable):
    """The execution of a DesignWorkflow.

    Parameters
    ----------
    uid: str
        Unique identifier of the workflow execution
    project_id: str
        Unique identifier of the project that contains the workflow execution
    workflow_id: str
        Unique identifier of the workflow that was executed
    version_number: int
        Integer identifier that increases each time the workflow is executed. The first execution
        has version_number = 1.

    """

    _response_key = None
    _paginator: Paginator = Paginator()
    _collection_key = 'response'

    uid: UUID = properties.UUID('id', serializable=False)
    workflow_id = properties.UUID('workflow_id', serializable=False)
    version_number = properties.Integer("version_number", serializable=False)

    status = properties.Optional(properties.String(), 'status', serializable=False)
    status_description = properties.Optional(
        properties.String(), 'status_description', serializable=False)
    status_info = properties.Optional(
        properties.List(properties.String()),
        'status_info',
        serializable=False
    )
    experimental = properties.Boolean("experimental", serializable=False, default=True)
    experimental_reasons = properties.Optional(
        properties.List(properties.String()),
        'experimental_reasons',
        serializable=False
    )
    created_by = properties.Optional(properties.UUID, 'created_by', serializable=False)
    updated_by = properties.Optional(properties.UUID, 'updated_by', serializable=False)
    create_time = properties.Optional(properties.Datetime, 'create_time', serializable=False)
    update_time = properties.Optional(properties.Datetime, 'update_time', serializable=False)

    score = properties.Object(Score, 'score')
    descriptors = properties.List(properties.Object(Descriptor), 'descriptors')

    def __init__(self):
        """This shouldn't be called, but it defines members that are set elsewhere."""
        self.project_id: Optional[UUID] = None  # pragma: no cover
        self.session: Optional[Session] = None  # pragma: no cover

    def __str__(self):
        return '<DesignExecution {!r}>'.format(str(self.uid))

    def _path(self):
        return '/projects/{project_id}/design-workflows/{workflow_id}/executions/{execution_id}' \
            .format(project_id=self.project_id,
                    workflow_id=self.workflow_id,
                    execution_id=self.uid)

    @classmethod
    def _build_candidates(cls, subset_collection: Iterable[dict]) -> Iterable[DesignCandidate]:
        for candidate in subset_collection:
            yield DesignCandidate.build(candidate)

    def candidates(self,
                   page: Optional[int] = None,
                   per_page: int = 100,
                   ) -> Iterable[DesignCandidate]:
        """Fetch the Design Candidates for the particular execution, paginated."""
        path = self._path() + '/candidates'

        fetcher = partial(self._fetch_page, path=path)

        return self._paginator.paginate(page_fetcher=fetcher,
                                        collection_builder=self._build_candidates,
                                        page=page,
                                        per_page=per_page)


class DesignExecutionCollection(Collection["DesignExecution"]):
    """A collection of DesignExecutions."""

    _path_template = '/projects/{project_id}/design-workflows/{workflow_id}/executions'  # noqa
    _individual_key = None
    _collection_key = 'response'
    _resource = DesignExecution

    def __init__(self, *,
                 project_id: UUID,
                 session: Session,
                 workflow_id: Optional[UUID] = None):
        self.project_id: UUID = project_id
        self.session: Session = session
        self.workflow_id: UUID = workflow_id

    def build(self, data: dict) -> DesignExecution:
        """Build an individual DesignWorkflowExecution."""
        execution = DesignExecution.build(data)
        execution.session = self.session
        execution.project_id = self.project_id
        return execution

    def trigger(self, execution_input: Score):
        """Trigger a Design Workflow execution given a score."""
        path = self._get_path()
        data = self.session.post_resource(path, {'score': execution_input.dump()})
        self._check_experimental(data)
        return self.build(data)

    def register(self, model: DesignExecution) -> DesignExecution:
        """Cannot register an execution."""
        raise NotImplementedError("Cannot register a DesignExecution.")

    def update(self, model: DesignExecution) -> DesignExecution:
        """Cannot update an execution."""
        raise NotImplementedError("Cannot update a DesignExecution.")

    def archive(self, execution_id: UUID):
        """Archive a Design Workflow execution.

        Parameters
        ----------
        execution_id: UUID
            Unique identifier of the execution to archive

        """
        raise NotImplementedError(
            "Design Executions cannot be archived")

    def restore(self, execution_id: UUID):
        """Restore an archived Design Workflow execution.

        Parameters
        ----------
        execution_id: UUID
            Unique identifier of the execution to restore

        """
        raise NotImplementedError(
            "Design Executions cannot be restored")

    def list(self,
             page: Optional[int] = None,
             per_page: int = 100,
             ) -> Iterable[DesignExecution]:
        """
        Paginate over the elements of the collection.

        Leaving page and per_page as default values will yield all elements in the
        collection, paginating over all available pages.

        Parameters
        ---------
        page: int, optional
            The "page" of results to list. Default is to read all pages and yield
            all results.  This option is deprecated.
        per_page: int, optional
            Max number of results to return per page. Default is 100.  This parameter
            is used when making requests to the backend service.  If the page parameter
            is specified it limits the maximum number of elements in the response.
        predictor_id: uuid, optional
            list executions that targeted the predictor with this id

        Returns
        -------
        Iterable[ResourceType]
            Resources in this collection.

        """
        return self._paginator.paginate(page_fetcher=self._fetch_page,
                                        collection_builder=self._build_collection_elements,
                                        page=page,
                                        per_page=per_page)

    def delete(self, uid: Union[UUID, str]) -> Response:
        """Design Workflow Executions cannot be deleted or archived."""
        raise NotImplementedError(
            "Design Executions cannot be deleted")



import time
from pprint import pprint
from citrine._rest.collection import Collection
from citrine.informatics.modules import Module
from citrine.resources.predictor_evaluation_execution import PredictorEvaluationExecution
from citrine.resources.design_execution import DesignExecution
from citrine.resources.workflow_executions import (
    WorkflowExecution,
    WorkflowExecutionStatus,
)
from typing import Union


class ConditionTimeoutError(RuntimeError):
    """Error that is raised when timeout is reached but the checked condition is still False."""

    pass


def _print_string_status(
    status: str, start_time: float, line_start: str = "", line_end: str = "\r"
):
    print(
        "{}Status = {:<25}Elapsed time".format(line_start, status),
        " = {}s".format(str(int(time.time() - start_time)).rjust(3)),
        end=line_end,
    )


def _pretty_execution_status(status: WorkflowExecutionStatus):
    status_text = status.status
    output_text = status_text if status_text != "InProgress" else "In progress"
    return output_text


def _print_execution_status(
    status: WorkflowExecutionStatus, start_time: float, line_end: str = "\r"
):
    print(
        "Status = {:<25}Elapsed time".format(_pretty_execution_status(status)),
        " = {}s".format(str(int(time.time() - start_time)).rjust(3)),
        end=line_end,
    )


def wait_while_validating(
    collection: Collection[Module],
    module: Module,
    print_status_info: bool = False,
    timeout: float = 1800.0,
    interval: float = 3.0,
) -> Module:
    """
    Wait until module is validated.

    Parameters
    ----------
    collection : Collection[Module]
        Collection containing module
    module : Module
        Module in Collection
    print_status_info : bool, optional
        Whether to print status info, by default False
    timeout : float, optional
        Maximum time spent inquiring in seconds, by default 1800.0
    interval : float, optional
        Inquiry interval in seconds, by default 3.0

    Returns
    -------
    Module
        Module in Collection after validation

    Raises
    ------
    ConditionTimeoutError
        If fails to validate within timeout

    """
    start = time.time()

    def is_validated():
        status = collection.get(module.uid).status
        if print_status_info:
            _print_string_status(status, start)
        return status != "VALIDATING" and status != "INPROGRESS"

    while not is_validated() and time.time() - start < timeout:
        time.sleep(interval)
    if not is_validated():
        msg = "Timeout reached, but condition is still {}".format(is_validated())
        raise ConditionTimeoutError(msg)

    if print_status_info:
        print("\nStatus info:")
        status_info = collection.get(module.uid).status_info
        pprint(status_info)

    return collection.get(module.uid)


def wait_while_executing(
    execution: Union[WorkflowExecution, PredictorEvaluationExecution, DesignExecution],
    print_status_info: bool = False,
    timeout: float = 1800.0,
    interval: float = 3.0,
    collection: Collection[Module] = None,
) -> Union[WorkflowExecution, PredictorEvaluationExecution, DesignExecution]:
    """
    Wait until execution is finished.

    Parameters
    ----------
    execution : WorkflowExecution, PredictorEvaluationExecution or DesignExecution
        an execution to monitor
    print_status_info : bool, optional
        Whether to print status info, by default False
    timeout : float, optional
        Maximum time spent inquiring in seconds, by default 1800.0
    interval : float, optional
        Inquiry interval in seconds, by default 3.0
    collection : Collection[Module]
        Collection containing module, only needed for PredictorEvaluationExecutions

    Returns
    -------
    WorkflowExecution, PredictorEvaluationExecution or DesignExecution
        the updated execution after it has finished executing


    Raises
    ------
    ConditionTimeoutError
        If fails to finish execution within timeout

    """
    start = time.time()

    def execution_is_finished():
        if isinstance(execution, (PredictorEvaluationExecution, DesignExecution)):
            if collection is None:
                raise ValueError("Must provide collection")
            status = collection.get(execution.uid).status
            if print_status_info:
                _print_string_status(status, start)
            return status != "INPROGRESS"
        else:
            status = execution.status()
            if print_status_info:
                _print_execution_status(status, start)
            return not status.in_progress

    while not execution_is_finished() and (time.time() - start < timeout):
        time.sleep(interval)
    if not execution_is_finished():
        msg = "Timeout reached, but condition is still {}".format(execution_is_finished())
        raise ConditionTimeoutError(msg)

    # re-fetch the execution if we have a collection to fetch it with
    if collection is not None:
        execution = collection.get(execution.uid)
    return execution



"""Resources that represent collections of predictors."""
from uuid import UUID
from typing import TypeVar, Optional

from citrine._session import Session
from citrine.resources.module import AbstractModuleCollection
from citrine.informatics.data_sources import DataSource
from citrine.informatics.predictors import Predictor, GraphPredictor

CreationType = TypeVar('CreationType', bound=Predictor)


class PredictorCollection(AbstractModuleCollection[Predictor]):
    """Represents the collection of all predictors for a project.

    Parameters
    ----------
    project_id: UUID
        the UUID of the project

    """

    _path_template = '/projects/{project_id}/modules'
    _individual_key = None
    _resource = Predictor
    _module_type = 'PREDICTOR'

    def __init__(self, project_id: UUID, session: Session):
        self.project_id = project_id
        self.session: Session = session

    def build(self, data: dict) -> Predictor:
        """Build an individual Predictor."""
        predictor: Predictor = Predictor.build(data)
        predictor._session = self.session
        predictor._project_id = self.project_id
        return predictor

    def check_for_update(self, predictor_id: UUID) -> Optional[Predictor]:
        """
        Check if there are updates available for a predictor.

        This check does not update the predictor; it just returns the update that is available.
        To perform the update, the response should then be used to call PredictorCollection.update

        Parameters
        ----------
        predictor_id: UUID
            Unique identifier of the predictor to check

        Returns
        -------
        Optional[Predictor]
            The update, if an update is available; None otherwise.

        """
        path = "/projects/{}/predictors/{}/check-for-update".format(self.project_id, predictor_id)
        data = self.session.get_resource(path)
        if data["updatable"]:
            enveloped = GraphPredictor.stuff_predictor_into_envelope(data["update"])
            built: Predictor = Predictor.build(enveloped)
            built.uid = predictor_id
            return built
        else:
            return None

    def auto_configure(self, training_data: DataSource, pattern="PLAIN") -> Predictor:
        """[ALPHA] Create a default predictor for some training data.

        This method will return an unregistered predictor generated by inspecting the
        training data and attempting to automatically configure the predictor.
        The automatic configuration tries to be as inclusive as possible.

        Parameters
        ----------
        training_data: DataSource
            The data to configure the predictor to model.
        pattern: str
            The predictor pattern to use, either "PLAIN" or "FORMULATION"

        Returns
        -------
        Predictor
            Automatically configured predictor for the training data

        """
        if pattern not in {"FORMULATION", "PLAIN"}:
            msg = "Called with pattern: {}.  Expected 'FORMULATION' or 'PLAIN'.".format(pattern)
            raise ValueError(msg)
        path = f'projects/{self.project_id}/predictors/default-predictor'
        body = {"data_source": training_data.dump(), "pattern": pattern}
        data = self.session.post_resource(path, json=body)
        if 'instance' in data:
            data['config'] = data.pop('instance')
        return self.build(data)
